#!/usr/bin/perl

use strict;
use warnings;

use File::Find qw(find);
use File::Spec::Functions qw(rel2abs);
use File::Temp qw(tempfile);
use Getopt::ArgvFile home=>1;
use Getopt::Long qw(GetOptions);
use HTTP::Request::Common qw(POST);
use LWP::UserAgent;
use WebService::MusicBrainz::Track;

# name of this program
my $progname = 'puidtool';
my $version = '0.6';

# command line options (defaulted)
my $confirm = 1; # true
my $generate = 1; # true
my $genpuid = 'genpuid';
my $incremental = 1; # true
my $lookup = 1; # true
my $mbrealm = 'musicbrainz.org';
my $mbserver = 'musicbrainz.org';
my $puidtag = 'MUSICIP_PUID';
my $quiet; # false
my $regen; # false
my $submit = 1; # true
my $trackidtag = 'MUSICBRAINZ_TRACKID';
my $verbose; # false
my $write = 1; # true

# command line options (undefined)
my $dnskey;
my $flac_dir;
my $mbpass;
my $mbuser;

# statistics
my $puids_generated = 0;
my $puids_found_in_tracks = 0;
my $puids_found_in_musicbrainz = 0;
my $puids_written = 0;
my $puids_submitted = 0;
my $tracks_with_no_puids = 0;
my $releases_submitted = 0;
my $total_tracks = 0;
my $tracks_without_trackid = 0;

# puid changes to queue up and send to musicbrainz
my %queue = ();

# cataloged files
my @files = ();

# hash of puid keys, each with array of filenames
my %puid_files = ();

# has a value if header has been output
my $header; # false

# returns "no" if boolean false (undefined)
sub noboolean {
  return (!defined($_[0]) ? 'no' : '');
}

sub not_set {
  return (defined($_[0]) ? $_[0] : 'not set');
}

# trims spaces from start and end of string
sub trim {
  return undef if (!defined($_[0]));
  my $v = $_[0];
  $v =~ s/^[[:space:]]+//; $v =~ s/[[:space:]]+$//;
  return $v;
}

sub trail_slash {
  return undef if (!defined($_[0]));
  my $v = $_[0];
  $v =~ s/([^\/])$/$1\//;
 return $_[0];
}

sub header {

  # do nothing if already output or user asked to be quiet
  return if (defined($header) || defined($quiet));

  print STDERR <<EOF;

$progname $version - Generate & reconcile PUID tags in FLAC files.
Copyright 2009 Paul C. Bryan. All rights reserved.

EOF

  $header = 1;
}

# display usage and exit
sub usage {

  undef $quiet;
  header();

	print STDERR <<EOF;
Usage: $progname [options] flac_dir

OPTIONS:
  --[no]confirm       Confirm PUIDs to submit prior to submission [${\(noboolean($confirm))}confirm].
  --dnskey=<key>      Key to query MusicDNS [${\(not_set($dnskey))}].
  --[no]generate      Generate PUID for track if no PUID tag found in track [${\(noboolean($generate))}generate].
  --genpuid=<prog>    Program to call to generate PUID of FLAC file [$genpuid].
  --[no]incremental   Perform incremental reconciliation with MusicBrainz [${\(noboolean($incremental))}incremental].
  --help              Display this help message.
  --[no]lookup        Perform lookup in MusicBrainz [${\(noboolean($lookup))}lookup].
  --mbpass=<passwd>   Password to authenticate with MusicBrainz.
  --mbrealm=<realm>   HTTP basic authentication realm for authentication [$mbrealm].
  --mbserver=<svr>    MusicBrainz server to query [$mbserver].
  --mbuser=<user>     Username to authenticate with MusicBrainz [${\(not_set($mbuser))}].
  --puidtag=<tag>     Name of tag in FLAC file to store PUID [$puidtag].
  --[no]quiet         Quiet operation, no screen output [${\(noboolean($quiet))}quiet].
  --[no]regen         Force regeneration of PUID [${\(noboolean($regen))}regen].
  --[no]submit        Submit PUID differences with MusicBrainz [${\(noboolean($submit))}submit].
  --trackidtag=<tag>  Name of tag in FLAC file to store track ID [$trackidtag].
  --[no]verbose       Verbose output of progress information [${\(noboolean($verbose))}verbose].
  --[no]write         Write PUID to FLAC file [${\(noboolean($write))}write].

PUID GENERATION
  If a track has no $puidtag tag, then its PUID will be generated by calling
  the $genpuid utility. If a track already has a $puidtag tag, it will be
  presumed correct and will not be regenerated, unless forced by the --regen
  option. If generation of the PUID differs from the track's PUID, then the
  generated PUID will be  written to the FLAC file, unless --nowrite is
  specified.

PUID RECONCILATION
  By default, this utility attempts to reconcile the PUID generated for the
  track with the track in MusicBrainz. It can only do this if the track has a
  $trackidtag tag that is stored in MusicBrainz. If a difference in the PUID
  is found, then the difference is submitted, unless --nosubmit is specified.
  Submissions for all specified tracks are batched in a single request.
  Incremental reconciliation only looks up PUIDs that are generated or changed
  based on PUID generation. Disabiling incremental reconciliation with the
  --noincremental option causes every track with a PUID to be compared to
  MusicBrainz.

CONFIGURATION FILE:
  You can store persistent command-line options in a file called .$progname in
  your home directory. They will be read as defaults prior to processing any
  passed command-line options.

EOF
	exit 1;
}

# output informative message if not quiet
sub msg {
  header(); print STDERR "$_[0]\n" if (!defined($quiet));
}

# display error message if not quiet, then exit with error code
sub error {
  undef $quiet; msg("$_[0]\n"); exit 1;
}

# output diagnostic if verbosity enabled
sub diag {
  msg($_[0]) if (defined($verbose));
}

# display system call as diagnostic, then execute system call
sub sys {
  diag($_[0]); system($_[0]) == 0 or error("Error calling $_[0].");
}

# return true if value defined and not empty
sub is_value {
  return (defined($_[0]) && length($_[0]) > 0);
}

# display (optional) warning message and prompt for user confirmation
sub warning {
  msg($_[0]);
  print STDERR "Press Enter to continue or ^C to cancel: ";
  (<STDIN>);
}

sub get_options {

  # parse command line options (may override defaults or initialization)
  GetOptions(
    'confirm'       => sub { $confirm = 1; },
    'dnskey=s'      => \$dnskey,
    'genpuid=s'     => \$genpuid,
    'generate'      => sub { $generate = 1; },
    'help'          => sub { usage(); },
    'incremental'   => sub { $incremental = 1; },
    'lookup'        => sub { $lookup = 1; },
    'mbuser=s'      => \$mbuser,
    'mbpass=s'      => \$mbpass,
    'mbrealm=s'     => \$mbrealm,
    'mbserver=s'    => \$mbserver,
    'noconfirm'     => sub { undef $confirm; },
    'nogenerate'    => sub { undef $generate; },
    'noincremental' => sub { undef $generate; },
    'nolookup'      => sub { undef $lookup; },
    'noregen'       => sub { undef $regen; },
    'nosubmit'      => sub { undef $submit; },
    'nowrite'       => sub { undef $write; },
    'puidtag=s'     => \$puidtag,
    'quiet'         => sub { $quiet = 1; undef $verbose; },
    'regen'         => sub { $regen = 1; },
    'submit'        => sub { $submit = 1; },
    'trackidtag=s'  => \$trackidtag,
    'verbose'       => sub { $verbose = 1; undef $quiet; },
    'write'         => sub { $write = 1; }
  ) or usage();

  # only one (optional) command-line argument
  usage if (scalar(@ARGV) != 1);

  # first parameter: directory to find flac files (make absolute)
  $flac_dir = rel2abs($ARGV[0]);

  # ensure director(ies) specified are indeed directories
  error("$flac_dir is not a directory.") if (!-d $flac_dir);

  # directories should have trailing slashes
  $flac_dir = trail_slash($flac_dir);
}

# escape special characters to make appropriate for quoted program arguments
sub escq {
  return undef if (!defined($_[0]));
  my $v = $_[0];
  $v =~ s/\"/\\\"/g;
  $v =~ s/\$/\\\$/g;
  return '"' . $v . '"';
}

# get metadata from flac file
sub get_metadata {
  my $flac = $_[0];
  my %metadata = ();
  open(METAFLAC, "metaflac --no-utf8-convert --export-tags-to=- " . escq($flac) . '|');
  while (<METAFLAC>) {
    chomp;
    utf8::encode($_);
    if (m/^(.*)=(.*)/) {
     $metadata{$1} = $2 if (is_value($1) && is_value($2));
    }
  }
  close(METAFLAC);
  return %metadata;
}

sub genpuid {

  return if (!$generate);

  error('The dnskey option must be specified to generate PUID.') if (!defined($dnskey));

  # get arguments
  my $file = $_[0];

  my $puid;

  open(PUID, "$genpuid $dnskey " . escq($file) . ' | grep "puid:" |');
  while (<PUID>) {
    chomp;
    if ($_ =~ /.*puid: (.*)/) {
      $puid = $1;
    }    
  }
  close(PUID);

  if (is_value($puid)) {
    $puids_generated++;
    diag("PUID generated: $puid.")
  }

  return $puid;
}

sub write_puid {

  return if (!$write);

  diag("Writing PUID to FLAC file.");
  sys("metaflac --remove-tag=$puidtag --set-tag=" . escq("$puidtag=$_[1]") . ' ' . escq($_[0]));
  $puids_written++;
}

# substituting with another subroutine for now because search has error
sub get_musicbrainz_track {

  my $tries = 3;
  my $interval = 1;

  while ($tries--) {
    sleep($interval);
    my $ws = WebService::MusicBrainz::Track->new();
    my $response;
    eval {
      $response = $ws->search({ MBID=>$_[0], INC=>'puids releases' });
    };
    return $response->track() if (!$@);
    $interval *= 2;
  }

  error("Get MusicBrainz track failed: $@");
}

sub lookup {

  return if (!$lookup);

  my $trackid = $_[0];
  my $puid = $_[1];

  # PUID must also be present to perform lookup
  return if (!is_value($trackid) || !is_value($puid));

  diag("Looking for matching PUID in MusicBrainz.");

  my $found;

  my $track = get_musicbrainz_track($trackid);
  my $releaseid = @{$track->release_list()->releases()}[0]->id();
  my $puid_list = $track->puid_list();

  if (is_value($puid_list)) {
    foreach my $mb_puid (@{$puid_list->puids()}) {
      if ($mb_puid->id() eq $puid) {
        $found = 1;
        last;
      }
    }
  }

  if ($found) {
    diag('Found in MusicBrainz.');
    $puids_found_in_musicbrainz++;
  }

  else {
    diag("PUID not in MusicBrainz" . ($submit ? '; queuing for submission' : '') . '.');
    $queue{$releaseid} = [] if (!defined($queue{$releaseid}));
    push(@{$queue{$releaseid}}, "$trackid $puid");
  }
}

sub process_file {

  msg("* $_[0]");

  # get arguments
  my $file = $_[0];

  my %metadata = get_metadata($file);
  my $trackid = $metadata{$trackidtag};

  # nothing to do if there is no track id associated with the track
  if (!is_value($trackid)) {
    diag("Track has no $trackidtag tag; ignoring.\n");
    $tracks_without_trackid++;
    return;
  }

  my $flac_puid = $metadata{$puidtag};
  my $puid = $flac_puid;

  if (is_value($flac_puid)) {
    diag("PUID in track: $flac_puid.");
    $puids_found_in_tracks++;
  }

  if (!is_value($puid) || $regen) {
	$puid = genpuid($file);
  }

  if (!is_value($puid)) {
    diag('PUID for file could not be established.');
    $tracks_with_no_puids++;
  }

  # determine if PUID changed from genpuid
  my $puid_changed = (is_value($puid) && (!is_value($flac_puid) || $puid ne $flac_puid));

  # write PUID changes to FLAC file
  if ($puid_changed) {
    write_puid($file, $puid);
  }

  # perform lookup in MB if PUID changed or full reconciliation requested
  if ($puid_changed || !$incremental) {
    lookup($trackid, $puid);
  }
}

sub push_file {

  # get the complete pathname to the file
  my $file = $_;

  # if file is a directory, ignore it
  return if (-d $file);

  # if file does not have flac extension, ignore it
  return if (!($file =~ m/\.flac$/));

  # file must be in flac directory because we're eventually going to compute relative name
  error("$file not in $flac_dir") if (index($file, $flac_dir) != 0);

  push(@files, $file);
}

sub process_files {

  msg('Cataloging files...');

  # get all files up-front
  find({wanted =>\&push_file, no_chdir=> 1, follow=>1, follow_skip=>2, dangling_symlinks=>0}, $flac_dir);

  # some deterministic way of iterating through the files found
  @files = sort { $a cmp $b } @files;

  msg('');
  msg('Processing files...');

  for my $file (@files) {
    process_file($file);
    $total_tracks++;
  }
}

# allow user to confirm processing before anything is written
sub confirm {

  # do nothing if explicitly told to be quiet
  return if (defined($quiet) || !$confirm);

  msg "Going to submit the following PUIDs:\n";

  foreach my $releaseid (keys(%queue)) {
    msg("Release ID: $releaseid");
    foreach my $submission (@{$queue{$releaseid}}) {
      msg "* $submission";
    }
    msg('');
  }

  warning("Ready to submit PUIDs to MusicBrainz.");
  msg('');
}

sub submit_puid_release {

  $releases_submitted++;

  my $tracks = shift;

  my $ua = LWP::UserAgent->new;

  my @form = ();
  push(@form, 'client'=>"$progname-$version");
  foreach my $track (@$tracks) {
    push(@form, 'puid'=>$track);
    $puids_submitted++;
  }

  $ua->agent("$progname/$version");
  $ua->credentials("$mbserver:80", $mbrealm, $mbuser=>$mbpass);

  my $req = POST("http://$mbserver/ws/1/track/", \@form);

  my $tries = 3;
  my $interval = 1;

  while ($tries--) {
    sleep($interval);
    my $response = $ua->request($req);
    return if ($response->code() == 200);
    $interval *= 2;
  }

  error("SUBMISSION FAILED");
}

sub submit {

  return if (!$submit);

  msg('');

  if (scalar(keys(%queue)) == 0) {
    msg("No items queued to submit to MusicBrainz.");
    return;
  }

  # user confirmation of submission
  confirm;

  msg("Submitting PUIDs to MusicBrainz...");

  foreach my $releaseid (keys(%queue)) {
    msg("Release ID: $releaseid.");
    submit_puid_release(\@{$queue{$releaseid}});
  }
}

sub statistics {

  msg("\nSTATISTICS\n");
  msg("* Total number of tracks:         $total_tracks");
  msg("* PUIDs found in track files:     $puids_found_in_tracks");
  msg("* PUIDs generated by genpuid:     $puids_generated");
  msg("* PUIDs couldn't be generated:    $tracks_with_no_puids");
  msg("* Tracks ignored (no trackid):    $tracks_without_trackid");
  msg("* PUIDs written to track files:   $puids_written");
  msg("* PUIDs found in MusicBrainz:     $puids_found_in_musicbrainz");
  msg("* PUIDs submitted to MusicBrainz: $puids_submitted");
  msg("* Releases for PUIDs submitted:   $releases_submitted");
}

# parse command line options
get_options();

# process the files specified
process_files();

# process queue of changes to submit to MusicBrainz
submit();

# output statistics
statistics();

msg('');
